#pragma once

#include "chainerx/array.h"

namespace chainerx {

Array Sinh(const Array& x);

Array Cosh(const Array& x);

Array Tanh(const Array& x);

Array Arcsinh(const Array& x);

Array Arccosh(const Array& x);

}  // namespace chainerx
